#ifndef OPTICALFLOW_H
#define OPTICALFLOW_H
#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
//#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/video/tracking.hpp>

#include <vector>
#include <time.h>
#include "FeatureSet.h"
//#include "FeatureExtraction.h"

class OpticalFlow
{
public:
	OpticalFlow(CvSize size);
	virtual ~OpticalFlow(void);

	int InitFlow(IplImage *image);
	int GetFlow2(IplImage *image, CvPoint2D32f *displacement_vector, CvPoint2D32f *velocity_vector, float *orientation);
	void ResetFlow();
	void DrawFlow(IplImage *out_image);
	int GetFASTFeatures(IplImage *image, CvPoint2D32f *pts);
	void ComputeFASTFeatures ( IplImage * image, FeatureSet & features );
	
private:
	IplImage *curr_image;
	IplImage *prev_image;

	IplImage *eig_image;
	IplImage *tmp_image;

	IplImage* pyr_prev;	//prev_image
	IplImage* pyr_curr;	//curr_image

	CvPoint2D32f *corners_prev;
	CvPoint2D32f *corners_curr;
	CvPoint2D32f *corners_output;
	CvSize winsize;

	time_t currtime;
	time_t prevtime;
	time_t starttime;
	double deltatime;

	int corner_count_prev;
	int corner_count_curr;
	typedef struct _matches matches_t;
	std::vector<matches_t> matches;

	int RansacFlow(std::vector<matches_t> &m, CvPoint2D32f *displacement, float *orientation);
};
#endif