/*
 * Processor.cpp
 *
 *  Created on: Jun 13, 2010
 *      Author: ethan
 */

#include "Processor.h"
#include "OpticalFlow.h"

#include <sys/stat.h>

#define RATIO_TH 0.4
#define SSD_TH 0.5
#define MOPS_WINDOW_SIZE 8
#define ALPH_W 30
#define FOCAL 686.06
#define IMG_W 640
#define IMG_H 480
#define FAST_TH 25

#define MIN_OPTICAL_PTS 100
#define MAX_ERROR 1000
#define MAX_DISP 150
#define FAILURE -1
#define SUCCESS 0
#define MAX_IMAGES 100
#define MIN_ELAPSETIME 1000

#include <vector>
#include <math.h>

#define MAX_CORNERS 200
#define WINSIZE 10
#define LEVELS 5
#define MAX_FEATURE_ERROR 550
#define MIN_MATCHES 30
#define USE_FAST 1
#define FAST_TH 25
#define DISP_ERROR_VAR 0.1	//pixel

struct _matches{
	CvPoint2D32f f1;
	CvPoint2D32f f2;
	float score;
};

enum HEADING{
	NORTH,
	NORTHEAST,
	EAST,
	SOUTHEAST,
	SOUTH,
	SOUTHWEST,
	WEST,
	NORTHWEST
};

enum CAM_MOTION
{
	CAM_MOVE_LEFT,
	CAM_MOVE_RIGHT
};

using namespace cv;

IplImage * constructPanorama ( IplImage *     img1,
                               IplImage *     img2,
                               int            featureType,
                               int            matchType, 
							   int * 		  width,
							   int *          height );
IplImage * compositeImages ( IplImage *     img1,
                             IplImage *     img2,
                             CvMat *        h,
							 int * 			width,
							 int *			height );
bool computeFeatures ( IplImage *     image,
                       FeatureSet &   features,
                       int featureType );
void ComputeSURFFeatures ( IplImage * image, FeatureSet & features );
void ComputeFASTFeatures ( IplImage * image, FeatureSet & features );
void ratioMatchFeatures ( const FeatureSet &     f1,
                          const FeatureSet &     f2,
                          vector<FeatureMatch> & matches,
                          float &                totalScore );
CvMat * ransacHomography ( const std::vector<Feature> &      f1,
                           const std::vector<Feature> &      f2,
                           const std::vector<FeatureMatch> & matches,
                           int                               numRounds,
                           int                               inlierThreshold );
void computeHomography ( const std::vector<Feature> &      f1,
                         const std::vector<Feature> &      f2,
                         const std::vector<FeatureMatch> & matches,
                         CvMat *                           h );
void applyHomography ( float   x,
                       float   y,
                       float & xNew,
                       float & yNew,
                       float   h [9] );
					   
void findPanoSize( IplImage * img1, IplImage * img2, CvMat * h, int * width, int * height);

float dx = 0.f;
float dy = 0.f;
float mid = 0.f;
  int img1atlhs = 1;
  
  float alph = 0.f;
  float alph1 = 1.f-alph;
					   
// Create a feature set.
FeatureSet::FeatureSet()
{
}

// Create a feature.
Feature::Feature ()
{
  selected = false;
}


Processor::Processor() :
			stard(20/*max_size*/, 8/*response_threshold*/,
					15/*line_threshold_projected*/,
					8/*line_threshold_binarized*/, 5/*suppress_nonmax_size*/),
			fastd(20/*threshold*/, true/*nonmax_suppression*/),
			surfd(100./*hessian_threshold*/, 1/*octaves*/, 2/*octave_layers*/)
{
}

Processor::~Processor() {
	// TODO Auto-generated destructor stub
	/*for (int i = 0; i < images.size(); ++i)
	{
		IplImage * junk = images[i];
		cvReleaseImage(&junk);
	}
	images.clear();*/
}

void Processor::detectAndDrawFeatures(int input_idx, image_pool* pool,int feature_type) {
	FeatureDetector* fd = 0;

	switch (feature_type) {
	case DETECT_SURF:
		fd = &surfd;
		break;
	case DETECT_FAST:
		fd = &fastd;
		break;
	case DETECT_STAR:
		fd = &stard;
		break;
	}

	Mat greyimage = pool->getGrey(input_idx);
	//Mat* grayimage = pool->getYUV(input_idx);

	Mat img = pool->getImage(input_idx);

	if (img.empty() || greyimage.empty() || fd == 0)
		return; //no image at input_idx!


	keypoints.clear();

	//if(grayimage->step1() > sizeof(uchar)) return;
	//cvtColor(*img,*grayimage,CV_RGB2GRAY);

	//const float f = FOCAL;
	//const float s = f;
	//const int offset_x = IMG_W/2;
	//const int offset_y = IMG_H/2;
	//double x, y;
	//int i, j, k;
	//unsigned char value;
	//int nChannels = 3;
	//IplImage * input = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_8U, nChannels);

	//for (int i = 0; i < IMG_W*IMG_H*nChannels; ++i)
	//{
	//	((uchar*)input->imageData)[i] = img->data[i];
	//}

	//for(j = -offset_y; j<IMG_H - offset_y; j++) 
	//{
	//	for(i = -offset_x; i<IMG_W - offset_x; i++)
	//	{
	//		//inverse mapping, this is the correct mapping
	//		x = f*tan(i/s) + offset_x;
	//		y = f*tan(j/s)*1/cos(i/s) + offset_y;
	//		
	//		//out of bounds
	//		if(x<0 || x>=IMG_W || y<0 || y>=IMG_H) 
	//		{
	//			img->data[(i + offset_x)*nChannels+(j + offset_y)*IMG_W*nChannels] = (uchar)0;
	//			img->data[(i + offset_x)*nChannels+(j + offset_y)*IMG_W*nChannels+1] = (uchar)0;
	//			img->data[(i + offset_x)*nChannels+(j + offset_y)*IMG_W*nChannels+2] = (uchar)0;
	//		}
	//		else 
	//		{
	//			for(k=0; k < nChannels; k++) 
	//			{
	//				value = abs((floor(y+1)-y))*abs((floor(x+1)-x))*CV_IMAGE_ELEM(input, uchar, (int)floor(y), ((int)floor(x))*nChannels + k) +
	//						abs((floor(y+1)-y))*abs((floor(x)-x))*CV_IMAGE_ELEM(input, uchar, (int)floor(y), ((int)ceil(x))*nChannels + k) +
	//						abs((floor(y)-y))*abs((floor(x+1)-x))*CV_IMAGE_ELEM(input, uchar, (int)ceil(y), ((int)floor(x))*nChannels + k) +
	//						abs((floor(y)-y))*abs((floor(x)-x))*CV_IMAGE_ELEM(input, uchar, (int)ceil(y), ((int)ceil(x))*nChannels + k);
	//				img->data[(i + offset_x)*nChannels + k +(j + offset_y)*IMG_W*nChannels] = value;
	//			}
	//		}
	//	}
	//}
	//cvReleaseImage(&input);

	fd->detect(greyimage, keypoints);

	for (vector<KeyPoint>::const_iterator it = keypoints.begin(); it
			!= keypoints.end(); ++it) {
		circle(img, it->pt, 3, cvScalar(255, 0, 255, 0));
	}
	//GaussianBlur(*img, *img, cvSize(5,5), 0);

	//pool->addImage(output_idx,outimage);

}
static double computeReprojectionErrors(
		const vector<vector<Point3f> >& objectPoints, const vector<vector<
				cv::Point2f> >& imagePoints, const vector<Mat>& rvecs,
		const vector<Mat>& tvecs, const Mat& cameraMatrix,
		const Mat& distCoeffs, vector<float>& perViewErrors) {
	vector<cv::Point2f> imagePoints2;
	int i, totalPoints = 0;
	double totalErr = 0, err;
	perViewErrors.resize(objectPoints.size());

	for (i = 0; i < (int) objectPoints.size(); i++) {
		projectPoints(Mat(objectPoints[i]), rvecs[i], tvecs[i], cameraMatrix,
				distCoeffs, imagePoints2);
		err = norm(Mat(imagePoints[i]), Mat(imagePoints2), CV_L1 );
		int n = (int) objectPoints[i].size();
		perViewErrors[i] = err / n;
		totalErr += err;
		totalPoints += n;
	}

	return totalErr / totalPoints;
}

static void calcChessboardCorners(Size boardSize, float squareSize, vector<
		Point3f>& corners) {
	corners.resize(0);

	for (int i = 0; i < boardSize.height; i++)
		for (int j = 0; j < boardSize.width; j++)
			corners.push_back(Point3f(float(j * squareSize), float(i
					* squareSize), 0));
}

/**from opencv/samples/cpp/calibration.cpp
 *
 */
static bool runCalibration(vector<vector<cv::Point2f> > imagePoints,
		Size imageSize, Size boardSize, float squareSize, float aspectRatio,
		int flags, Mat& cameraMatrix, Mat& distCoeffs, vector<Mat>& rvecs,
		vector<Mat>& tvecs, vector<float>& reprojErrs, double& totalAvgErr) {
	cameraMatrix = Mat::eye(3, 3, CV_64F);
	if (flags & CV_CALIB_FIX_ASPECT_RATIO)
		cameraMatrix.at<double> (0, 0) = aspectRatio;

	distCoeffs = Mat::zeros(5, 1, CV_64F);

	vector<vector<Point3f> > objectPoints(1);
	calcChessboardCorners(boardSize, squareSize, objectPoints[0]);
	for (size_t i = 1; i < imagePoints.size(); i++)
		objectPoints.push_back(objectPoints[0]);

	calibrateCamera(objectPoints, imagePoints, imageSize, cameraMatrix,
			distCoeffs, rvecs, tvecs, flags);

	bool ok = checkRange(cameraMatrix, CV_CHECK_QUIET ) && checkRange(
			distCoeffs, CV_CHECK_QUIET );

	totalAvgErr = computeReprojectionErrors(objectPoints, imagePoints, rvecs,
			tvecs, cameraMatrix, distCoeffs, reprojErrs);

	return ok;
}

bool Processor::detectAndDrawChessboard(int idx,image_pool* pool) {

	Mat grey = pool->getGrey(idx);
	if (grey.empty())
		return false;
	vector<cv::Point2f> corners;

	IplImage iplgrey = grey;
	if (!cvCheckChessboard(&iplgrey, Size(6, 8)))
		return false;
	bool patternfound = findChessboardCorners(grey, Size(6, 8), corners);

	Mat img = pool->getImage(idx);

	if (corners.size() < 1)
		return false;

	cornerSubPix(grey, corners, Size(11, 11), Size(-1, -1), TermCriteria(
			CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));

	if(patternfound)
		imagepoints.push_back(corners);

	drawChessboardCorners(img, Size(6, 8), Mat(corners), patternfound);

	imgsize = grey.size();

	return patternfound;

}

void Processor::drawText(int i, image_pool* pool, const char* ctext){
	// Use "y" to show that the baseLine is about
	string text = ctext;
	int fontFace = FONT_HERSHEY_COMPLEX_SMALL;
	double fontScale = .8;
	int thickness = .5;

	Mat img = pool->getImage(i);

	int baseline=0;
	Size textSize = getTextSize(text, fontFace,
	                            fontScale, thickness, &baseline);
	baseline += thickness;

	// center the text
	Point textOrg((img.cols - textSize.width)/2,
	              (img.rows - textSize.height *2));

	// draw the box
	rectangle(img, textOrg + Point(0, baseline),
	          textOrg + Point(textSize.width, -textSize.height),
	          Scalar(0,0,255),CV_FILLED);
	// ... and the baseline first
	line(img, textOrg + Point(0, thickness),
	     textOrg + Point(textSize.width, thickness),
	     Scalar(0, 0, 255));

	// then put the text itself
	putText(img, text, textOrg, fontFace, fontScale,
	        Scalar::all(255), thickness, 8);
}
void saveCameraParams(const string& filename, Size imageSize, Size boardSize,
		float squareSize, float aspectRatio, int flags,
		const Mat& cameraMatrix, const Mat& distCoeffs,
		const vector<Mat>& rvecs, const vector<Mat>& tvecs,
		const vector<float>& reprojErrs,
		const vector<vector<cv::Point2f> >& imagePoints, double totalAvgErr) {
	FileStorage fs(filename, FileStorage::WRITE);

	time_t t;
	time(&t);
	struct tm *t2 = localtime(&t);
	char buf[1024];
	strftime(buf, sizeof(buf) - 1, "%c", t2);

	fs << "calibration_time" << buf;

	if (!rvecs.empty() || !reprojErrs.empty())
		fs << "nframes" << (int) std::max(rvecs.size(), reprojErrs.size());
	fs << "image_width" << imageSize.width;
	fs << "image_height" << imageSize.height;
	fs << "board_width" << boardSize.width;
	fs << "board_height" << boardSize.height;
	fs << "squareSize" << squareSize;

	if (flags & CV_CALIB_FIX_ASPECT_RATIO)
		fs << "aspectRatio" << aspectRatio;

	if (flags != 0) {
		sprintf(buf, "flags: %s%s%s%s",
				flags & CV_CALIB_USE_INTRINSIC_GUESS ? "+use_intrinsic_guess"
						: "",
				flags & CV_CALIB_FIX_ASPECT_RATIO ? "+fix_aspectRatio" : "",
				flags & CV_CALIB_FIX_PRINCIPAL_POINT ? "+fix_principal_point"
						: "",
				flags & CV_CALIB_ZERO_TANGENT_DIST ? "+zero_tangent_dist" : "");
		cvWriteComment(*fs, buf, 0);
	}

	fs << "flags" << flags;

	fs << "camera_matrix" << cameraMatrix;
	fs << "distortion_coefficients" << distCoeffs;

	fs << "avg_reprojection_error" << totalAvgErr;
	if (!reprojErrs.empty())
		fs << "per_view_reprojection_errors" << Mat(reprojErrs);

	if (!rvecs.empty() && !tvecs.empty()) {
		Mat bigmat(rvecs.size(), 6, CV_32F);
		for (size_t i = 0; i < rvecs.size(); i++) {
			Mat r = bigmat(Range(i, i + 1), Range(0, 3));
			Mat t = bigmat(Range(i, i + 1), Range(3, 6));
			rvecs[i].copyTo(r);
			tvecs[i].copyTo(t);
		}
		cvWriteComment(
				*fs,
				"a set of 6-tuples (rotation vector + translation vector) for each view",
				0);
		fs << "extrinsic_parameters" << bigmat;
	}

	if (!imagePoints.empty()) {
		Mat imagePtMat(imagePoints.size(), imagePoints[0].size(), CV_32FC2);
		for (size_t i = 0; i < imagePoints.size(); i++) {
			Mat r = imagePtMat.row(i).reshape(2, imagePtMat.cols);
			Mat(imagePoints[i]).copyTo(r);
		}
		fs << "image_points" << imagePtMat;
	}
}
void Processor::resetChess() {

	imagepoints.clear();
}

void Processor::calibrate(const char* filename) {

	vector<Mat> rvecs, tvecs;
	vector<float> reprojErrs;
	double totalAvgErr = 0;
	int flags = 0;
	bool writeExtrinsics = true;
	bool writePoints = true;

	bool ok = runCalibration(imagepoints, imgsize, Size(6, 8), 1.f, 1.f,
			flags, K, distortion, rvecs, tvecs, reprojErrs, totalAvgErr);


	if (ok){

		saveCameraParams(filename, imgsize, Size(6, 8), 1.f,
				1.f, flags, K, distortion, writeExtrinsics ? rvecs
						: vector<Mat> (), writeExtrinsics ? tvecs
						: vector<Mat> (), writeExtrinsics ? reprojErrs
						: vector<float> (), writePoints ? imagepoints : vector<
						vector<cv::Point2f> > (), totalAvgErr);
	}

}

OpticalFlow * opflow;
time_t starttime;
time_t currtime;
time_t prevtime;
time_t elaspsedtime;
float error=0;
CvPoint2D32f displacement = cvPoint2D32f(0,0);
CvPoint2D32f velocity = cvPoint2D32f(0,0);
float orientation=0;
CvPoint2D32f accum_dist=cvPoint2D32f(0,0);
CvPoint2D32f d_accum_dist=cvPoint2D32f(0,0);
int img_taken[8] = {0};
int direction = CAM_MOVE_RIGHT;
void Processor::opticalFlowInit(image_pool* pool)
{
	Mat img = pool->getImage(0);
	IplImage * gray = cvCreateImage( cvSize(IMG_W, IMG_H), IPL_DEPTH_8U, 1);
	for (int i = 0; i < IMG_W*IMG_H; ++i)
	{
		gray->imageData[i] = img.data[i*3+1];
	}
	opflow = new OpticalFlow(cvSize(IMG_W,IMG_H));

	int status = 0;

	prevtime=currtime=clock();
	status = opflow->InitFlow(gray);
	cvReleaseImage(&gray);
	setSphericalCoord(FOCAL, FOCAL);
}

void Processor::opticalFlow(image_pool* pool, int* capture, int* distx, int* disty)
{
	Mat img = pool->getImage(0);
	IplImage * gray = cvCreateImage( cvSize(IMG_W, IMG_H), IPL_DEPTH_8U, 1);
	for (int i = 0; i < IMG_W*IMG_H; ++i)
	{
		gray->imageData[i] = img.data[i*3+1];
	}
	
	int status = 0;
			
	//Run Optical Flow
	currtime = clock();
	
	status = opflow->GetFlow2(gray, &displacement, &velocity, &orientation);
	
	//get accumulative displacement
	accum_dist.x += displacement.x;
	accum_dist.y += displacement.y;
	d_accum_dist.x += displacement.x;
	d_accum_dist.y += displacement.y;
	*distx = accum_dist.x;
	*disty = accum_dist.y;
	
	if(status != FAILURE){
		elaspsedtime = (currtime-prevtime);	//millisecond
		int x_region = accum_dist.x/MAX_DISP;
		if (x_region == -1 && img_taken[0] == 0)
		{
			direction = CAM_MOVE_LEFT;
		}
		if (direction == CAM_MOVE_LEFT)
		{
			x_region *= -1;
		}
		if (x_region < 0)
			return;
			
		//int y_region = accum_dist.y/MAX_DISP;
		if( abs(d_accum_dist.x)>MAX_DISP && !img_taken[x_region]){
			prevtime=currtime;
			d_accum_dist=cvPoint2D32f(0,0);
			img_taken[x_region] = 1;
			accum_offset.push_back(accum_dist);
			*capture = 1;
		}
	}
	cvReleaseImage(&gray);
}

void Processor::saveImage(int input_idx,image_pool* pool, char* BYTE)
{
	Mat img = pool->getImage(input_idx);

	if (img.empty())
		return; //no image at input_idx!

	for (int i = 0; i < img.rows*img.cols*3; ++i)
	{
		BYTE[i] = (char) img.data[i];//CV_MAT_ELEM(*img, char, i, j);
	}

}

int Processor::getNumberDetectedChessboards() {
	return imagepoints.size();
}

void Processor::captureImage(int input_idx, image_pool* pool, int output_idx)
{
	pool->addImage(output_idx, pool->getImage(input_idx));
}

double  sphericalCoordx[IMG_W][IMG_H];
double  sphericalCoordy[IMG_W][IMG_H];

void Processor::setSphericalCoord(float f, float s)
{
	int x, y;

	for(x=-IMG_W/2; x<IMG_W - IMG_W/2; x++) {
		for(y=-IMG_H/2; y<IMG_H - IMG_H/2; y++) {
			sphericalCoordx[x+IMG_W/2][y+IMG_H/2] = f*tan(x/s) + IMG_W/2;
			sphericalCoordy[x+IMG_W/2][y+IMG_H/2] = f*tan(y/s)*1/cos(x/s) + IMG_H/2;
			if((0>sphericalCoordx[x+IMG_W/2][y+IMG_H/2]) || (IMG_W<= sphericalCoordx[x+IMG_W/2][y+IMG_H/2]) ||
					(0>sphericalCoordy[x+IMG_W/2][y+IMG_H/2]) || (IMG_H<= sphericalCoordy[x+IMG_W/2][y+IMG_H/2])) {
				sphericalCoordx[x+IMG_W/2][y+IMG_H/2] = -1;
				sphericalCoordy[x+IMG_W/2][y+IMG_H/2] = -1;
			}
		}
	}
	return;
}

void Processor::getSphericalImage(char* output)
{
	const float f = FOCAL;
	const float s = f;
	double orig_x, orig_y;
	int sph_x, sph_y, img_ch, sph_coord, orig_coord[4];
	unsigned char value, v1, v2, v3, v4;
	int nChannels = 3;
	char input[IMG_W*IMG_H*nChannels];
	/*IplImage * input = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_8U, nChannels);
	for (int i = 0; i < IMG_W*IMG_H*nChannels; ++i)
	{
		((uchar*)input->imageData)[i] = (uchar) output[i];
	}*/
	memcpy(input, output, IMG_W*IMG_H*nChannels);

	for(sph_y = 0; sph_y<IMG_H; sph_y++) 
	{
		for(sph_x = 0; sph_x<IMG_W; sph_x++)
		{
			//inverse mapping, this is the correct mapping
			//orig_x = f*tan(sph_x/s) + offset_x;
//			orig_y = f*tan(sph_y/s)*1/cos(sph_y/s) + offset_y;

			orig_x=sphericalCoordx[sph_x][sph_y];
			orig_y=sphericalCoordy[sph_x][sph_y];
			sph_coord = ((sph_x)+(sph_y)*IMG_W)*nChannels;

			//out of bounds
			if(-1 == sphericalCoordx[sph_x][sph_y]) 
			{
				output[sph_coord+0] = (uchar)0;
				output[sph_coord+1] = (uchar)0;
				output[sph_coord+2] = (uchar)0;
			}
			else 
			{
				v1 = abs((floor(orig_y+1)-orig_y))*abs((floor(orig_x+1)-orig_x));
				v2 = abs((floor(orig_y+1)-orig_y))*abs((floor(orig_x)-orig_x));
				v3 = abs((floor(orig_y)-orig_y))*abs((floor(orig_x+1)-orig_x));
				v4 = abs((floor(orig_y)-orig_y))*abs((floor(orig_x)-orig_x));
				orig_coord[0] = ((floor(orig_x))+(floor(orig_y))*IMG_W)*nChannels;
				orig_coord[1] = ((ceil(orig_x))+(floor(orig_y))*IMG_W)*nChannels;
				orig_coord[2] = ((floor(orig_x))+(ceil(orig_y))*IMG_W)*nChannels;
				orig_coord[3] = ((ceil(orig_x))+(ceil(orig_y))*IMG_W)*nChannels;
				for(img_ch=0; img_ch < nChannels; img_ch++) 
				{
					value = v1*input[orig_coord[0]+img_ch] + //CV_IMAGE_ELEM(input, uchar, (int), ((int))*nChannels + img_ch) +
							v2*input[orig_coord[1]+img_ch] + //CV_IMAGE_ELEM(input, uchar, (int), ((int))*nChannels + img_ch) +
							v3*input[orig_coord[2]+img_ch] + //CV_IMAGE_ELEM(input, uchar, (int), ((int))*nChannels + img_ch) +
							v4*input[orig_coord[3]+img_ch]; //CV_IMAGE_ELEM(input, uchar, (int), ((int))*nChannels + img_ch);
					output[sph_coord + img_ch] = value;
				}
			}
		}
	}
	//cvReleaseImage(&input);
}

IplImage * pano;
void Processor::makePanorama1(char* img1c, char* img2c, image_pool* pool, int* width, int* height)
{
	IplImage * fimg1 = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_32F, 3);
	IplImage * fimg2 = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_32F, 3);
	for (int i = 0; i < IMG_H*IMG_W*3; i+=3)
	{
		((float*)fimg1->imageData)[i] = ((float) img1c[i]/255.);
		((float*)fimg1->imageData)[i+1] = ((float) img1c[i+1]/255.);
		((float*)fimg1->imageData)[i+2] = ((float) img1c[i+2]/255.);

		((float*)fimg2->imageData)[i] = ((float) img2c[i]/255.);
		((float*)fimg2->imageData)[i+1] = ((float) img2c[i+1]/255.);
		((float*)fimg2->imageData)[i+2] = ((float) img2c[i+2]/255.);
	}
 
	pano = constructPanorama( fimg1, fimg2, 2, 2 , width, height);
	//cvConvertScale(result, result, 255, 0);
	
	cvReleaseImage(&fimg1);
	cvReleaseImage(&fimg2);
}

void Processor::makePanorama2(char* img1c, image_pool* pool, int width, int height)
{
	for (int i = 0; i < width*height*3; ++i)
		img1c[i] = (char) (((float*)pano->imageData)[i]*255.);

	cvReleaseImage(&pano);
}

void Processor::makePanoUseOpticalFlow1(char* img1c, char* img2c, int* width, int* height)
{
	float xwant = atan((accum_offset[1].x-accum_offset[0].x)/FOCAL)*FOCAL;
	float ywant = atan((accum_offset[1].y-accum_offset[0].y)/FOCAL*cos(xwant/FOCAL))*FOCAL;
	float f_h[] = {	1, 0, xwant,
					0, 1, ywant,
					0, 0, 1 };
	CvMat * h = cvCreateMat(3, 3, CV_32F);
	for (int i = 0; i < 3; ++i)
	{
		for (int j = 0; j < 3; ++j)
		{
			CV_MAT_ELEM(*h, float, i, j) = f_h[i*3+j];
		}
	}
	IplImage * fimg1 = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_32F, 3);
	IplImage * fimg2 = cvCreateImage( cvSize( IMG_W,IMG_H ), IPL_DEPTH_32F, 3);
	for (int i = 0; i < IMG_H*IMG_W*3; i+=3)
	{
		((float*)fimg1->imageData)[i] = ((float) img1c[i]/255.);
		((float*)fimg1->imageData)[i+1] = ((float) img1c[i+1]/255.);
		((float*)fimg1->imageData)[i+2] = ((float) img1c[i+2]/255.);

		((float*)fimg2->imageData)[i] = ((float) img2c[i]/255.);
		((float*)fimg2->imageData)[i+1] = ((float) img2c[i+1]/255.);
		((float*)fimg2->imageData)[i+2] = ((float) img2c[i+2]/255.);
	}
	//pano = constructPanorama( fimg1, fimg2, 2, 2 , width, height);
	findPanoSize(fimg1, fimg2, h, width, height);
	pano = compositeImages(fimg1, fimg2, h, width, height);
	cvReleaseImage(&fimg1);
	cvReleaseImage(&fimg2);
	
}
void Processor::makePanoUseOpticalFlow2(char* img, int width, int height)
{
	for (int i = 0; i < width*height*3; ++i)
		img[i] = (char) (((float*)pano->imageData)[i]*255.);
	cvReleaseImage(&pano);
}

IplImage * constructPanorama ( IplImage *     img1,
                               IplImage *     img2,
                               int            featureType,
                               int            matchType,
							   int *		  width,
							   int * 		  height )
{
  assert ( img1->depth == IPL_DEPTH_32F );
  assert ( img1->nChannels == 3 );
  assert ( img2->depth == IPL_DEPTH_32F );
  assert ( img2->nChannels == 3 );

  int numRounds = 1000;
  int inlierThreshold = 5;
  FeatureSet f1, f2;
  vector<FeatureMatch> matches;
  float totalScore;
  printf("constructPanorama..\n");
  computeFeatures( img1, f1, featureType );
  computeFeatures( img2, f2, featureType );
  ratioMatchFeatures( f1, f2, matches, totalScore );
  printf("finished matching..\n");

  CvMat * h = ransacHomography( f1, f2, matches, numRounds, inlierThreshold );

  printf("found H..\n");
  findPanoSize(img1, img2, h, width, height);
  IplImage * result = compositeImages ( img1, img2, h , width, height);
  return result;

}

void findPanoSize( IplImage * img1, IplImage * img2, CvMat * h, int * width, int * height)
{
	vector<int> img21x;
  vector<int> img21y;
  float newx = 0;
  float newy = 0;
  float * float_h = h->data.fl;

  applyHomography( 0, 0, newx, newy, float_h );
  img21x.push_back((int)newx);
  img21y.push_back((int)newy);
  applyHomography( 0, img2->height-1, newx, newy, float_h );
  img21x.push_back((int)newx);
  img21y.push_back((int)newy);
  applyHomography( img2->width-1, 0, newx, newy, float_h );
  img21x.push_back((int)newx);
  img21y.push_back((int)newy);
  applyHomography( img2->width-1, img2->height-1, newx, newy, float_h );
  img21x.push_back((int)newx);
  img21y.push_back((int)newy);

  printf("Finding canvas size..\n");
  int xmin21 = *(std::min_element(img21x.begin(), img21x.end()));
  int xmax21 = *(std::max_element(img21x.begin(), img21x.end()));
  int ymin21 = *(std::min_element(img21y.begin(), img21y.end()));
  int ymax21 = *(std::max_element(img21y.begin(), img21y.end()));
  printf("xmin21 = %d xmax21 = %d ymin21 = %d ymax21 = %d\n",xmin21,xmax21,ymin21,ymax21);

  int xmin = min( 0, xmin21);
  int xmax = max( img1->width, xmax21);
  int ymin = min( 0, ymin21);
  int ymax = max( img1->height, ymax21);
  *width = xmax - xmin;
  *height = ymax - ymin;
  
  if ( xmin < 0.f )
      dx = -xmin;
  if ( ymin < 0.f )
      dy = -ymin;
  printf("width = %d \t height = %d\n",*width,*height);
  mid = (float)(img1->width+dx+img21x[0]+dx)/2.f;
  img1atlhs = 1;
  if (xmin21+xmax21 < img1->width+2*dx)
      img1atlhs = 0;
  alph = 0.f;
  alph1 = 1.f-alph;
}
IplImage * compositeImages ( IplImage *     img1,
                             IplImage *     img2,
                             CvMat *        h,  
							 int * 			width,
							 int *			height )
{
  /*assert ( img1->depth == IPL_DEPTH_32F );
  assert ( img1->nChannels == 3 );
  assert ( img2->depth == IPL_DEPTH_32F );
  assert ( img2->nChannels == 3 );*/
  float * float_h = h->data.fl;
  
  cvInvert(h, h, cv::DECOMP_SVD);
  int img1x = -1;
  int img1y = -1;
  int img2x = -1;
  int img2y = -1;
  float newx = 0;
  float newy = 0;
  
  

  IplImage *result = cvCreateImage(cvSize(*width, *height), IPL_DEPTH_32F, 3);
  for (int y = 0; y < result->height; ++y)
  {
      for (int x = 0; x < result->width; ++x)
      {
          alph = (x-mid)*abs(mid)/(result->width*ALPH_W);
          if (alph > 1.f)
              alph = 1.f;
          else if (alph < 0.f)
              alph = 0.f;
          if (img1atlhs)
              alph = 1-alph;
          alph1 = 1-alph;

          applyHomography( x-dx, y-dy, newx, newy, float_h );
          if ((int)newx < img2->width && (int)newx > 0 &&
              (int)newy < img2->height && (int)newy > 0 )
          {
              img2x = newx;
              img2y = newy;
          }
          else
          {
              img2x = -1;
              img2y = -1;
          }
          if ((x-(int)dx) < img1->width && (x-(int)dx) > 0 &&
              (y-(int)dy) < img1->height && (y-(int)dy) > 0)
          {
              img1x = (x-(int)dx);
              img1y = (y-(int)dy);
          }
          else
          {
              img1x = -1;
              img1y = -1;
          }

          for (int c = 0; c < result->nChannels; ++c)
          {
              if(img1x > 0)
              {
                  CV_IMAGE_ELEM(result, float, y, x*result->nChannels+c)
                  = CV_IMAGE_ELEM(img1, float, y-(int)dy, (x-(int)dx)*img1->nChannels+c);
                  if(img2x > 0)
                      CV_IMAGE_ELEM(result, float, y, x*result->nChannels+c)
                      = alph1*CV_IMAGE_ELEM(img2, float, (int)newy, (int)newx*img2->nChannels+c)
                      +alph*CV_IMAGE_ELEM(result, float, y, x*result->nChannels+c);
              }
              else if(img2x > 0)
                  CV_IMAGE_ELEM(result, float, y, x*result->nChannels+c)
                  = CV_IMAGE_ELEM(img2, float, (int)newy, (int)newx*img2->nChannels+c);

          }
      }
  }

  printf("done compositeImages");
  return result;
}

// Compute features of an image.
bool computeFeatures ( IplImage *     image,
                       FeatureSet &   features,
                       int featureType          )
{
  // Instead of calling dummyComputeFeatures, write your own
  // feature computation routines and call them here.
  switch ( featureType )
  {
	case 2:
    ComputeFASTFeatures ( image, features );
    break;
  case 3:
    ComputeSURFFeatures ( image, features );
    break;
  default:
    return false;
  }

  // This is just to make sure the IDs are assigned in order, because
  // the ID gets used to index into the feature array.
  for ( size_t i = 0; i < features.size (); ++i )
  {
    features [i].id = i + 1;
  }

  return true;
}

void ComputeFASTFeatures ( IplImage * image, FeatureSet & features )
{
	//cv::FastFeatureDetector fast;
	cv::SurfDescriptorExtractor surfdes;
	
	std::vector<cv::KeyPoint> keypoints;
	cv::Mat desM;
	//std::vector<float> descriptors;
	IplImage * temp = cvCreateImage ( cvGetSize ( image ), IPL_DEPTH_8U, 3 );
	cvConvertScale( image, temp, 255, 0 );
	IplImage * temp2 = cvCreateImage ( cvGetSize ( image ), IPL_DEPTH_8U, 1 );
	cvCvtColor ( temp, temp2, CV_RGB2GRAY );
	cvReleaseImage ( &temp );
	//fast.detect(temp2, keypoints);
	cv::FAST(temp2, keypoints, FAST_TH);
	surfdes.compute(temp2, keypoints, desM);
	cvReleaseImage ( &temp2 );
	features.clear ();
	features.reserve ( keypoints.size () );
	for ( size_t i = 0; i < keypoints.size (); ++i )
	{
		Feature f;
		f.x = keypoints[i].pt.x;
		f.y = keypoints[i].pt.y;
		f.type = 3;
		f.id = 0;
		f.angleRadians = keypoints[i].angle;
		f.data.reserve ( surfdes.descriptorSize () );
		for ( int j = 0; j < surfdes.descriptorSize (); ++j )
		{
			f.data.push_back ( desM.at<float>(i,j));// [i * surfdes.descriptorSize () + j] );
		}
		features.push_back ( f );
	}
	
}

void ComputeSURFFeatures ( IplImage * image, FeatureSet & features )
{
  cv::SURF surf;

  std::vector<cv::KeyPoint> keypoints;
  std::vector<float> descriptors;
  IplImage * temp = cvCreateImage ( cvGetSize ( image ), IPL_DEPTH_8U, 3 );
  cvConvertScale( image, temp, 255, 0 );
  IplImage * temp2 = cvCreateImage ( cvGetSize ( image ), IPL_DEPTH_8U, 1 );
  cvCvtColor ( temp, temp2, CV_RGB2GRAY );
  cvReleaseImage ( &temp );
  surf ( temp2, cv::Mat (), keypoints, descriptors );
  cvReleaseImage ( &temp2 );
  features.clear ();
  features.reserve ( keypoints.size () );
  for ( size_t i = 0; i < keypoints.size (); ++i )
  {
    Feature f;
    f.x = keypoints[i].pt.x;
    f.y = keypoints[i].pt.y;
    f.type = 3;
    f.id = 0;
    f.angleRadians = keypoints[i].angle;
    f.data.reserve ( surf.descriptorSize () );
    for ( int j = 0; j < surf.descriptorSize (); ++j )
    {
      f.data.push_back ( descriptors [i * surf.descriptorSize () + j] );
    }
    features.push_back ( f );
  }
}

void ratioMatchFeatures ( const FeatureSet &     f1,
                          const FeatureSet &     f2,
                          vector<FeatureMatch> & matches,
                          float &                totalScore )
{
  cv::Mat features ( f2.size (), f2[0].data.size (), CV_32F );
  int knn = 2;

  for ( size_t i = 0; i < f2.size (); ++i )
  {
    for ( size_t j = 0; j < f2[i].data.size (); ++j )
    {
      features.at<float>( ( int )i, ( int )j ) = f2[i].data[j];
    }
  }
  cv::flann::Index flannIndex ( features, cv::flann::KDTreeIndexParams () );
  // End FLANN index population

  totalScore = 0;
  float dBest = 1e100;
  float dBest2 = 1e100;
  int iBest = 0;

  for ( FeatureSet::const_iterator i1 = f1.begin (); i1 != f1.end (); ++i1 )
  {
    vector<int> indices;
    vector<float> dists;

    indices.resize(2);
    dists.resize(2);
    flannIndex.knnSearch(i1->data, indices, dists, knn, cv::flann::SearchParams());

    dBest = dists[0];
    dBest2 = dists[1];
    iBest = indices[0];

    FeatureMatch m;
    if (dBest/dBest2 < RATIO_TH)
    {
        m.score = dBest;  
        m.id1 = i1->id;
        m.id2 = f2[iBest].id;
        totalScore -= 1;
        matches.push_back ( m );
    }
  }
}

CvMat * ransacHomography ( const std::vector<Feature> &      f1,
                           const std::vector<Feature> &      f2,
                           const std::vector<FeatureMatch> & matches,
                           int                               numRounds,
                           int                               inlierThreshold )
{
  int rounds = 0;
  int count = 4;    // Pick only a random subset (4) of matches
  int best = 0;
  int bestinlierCount = 0;
  int inlierCount = 0;
  cv::Mat hmat( 9, 1, CV_32F);
  CvMat h = hmat;
  srand(5);
  int r = 0;
  float hbest[9] = {0};
  int c = 0;

  while(rounds++ < numRounds)
  {
      vector<FeatureMatch> submatches;
      inlierCount = 0;
      count = 4;

      while(count--)
      {
          r = (int)rand()%matches.size();
          while(matches[r].score >= SSD_TH)
		  {
              r = (int)rand()%matches.size();
			  if (++c > 20)
				  break;
		  }
		  if (c > 20)
			break;
          submatches.push_back(matches[r]);
      }
	  if (submatches.size() < 4)
		  continue;

      computeHomography( f1, f2, submatches, &h );
      float * float_h = h.data.fl;

      for ( int i = 0; i < matches.size(); ++i )
      {
          float xNew = 0.f;
          float yNew = 0.f;
          if (matches[i].score < SSD_TH)
          {
              applyHomography( f2[matches[i].id2-1].x, f2[matches[i].id2-1].y, xNew, yNew, float_h );

              // Find error
              int error = (int) pow( (double)(f1[matches[i].id1-1].x - xNew), (double)2.f ) + pow( (double)(f1[matches[i].id1-1].y - yNew), (double)2.f );

              if ( error < inlierThreshold*inlierThreshold )
              {
                  inlierCount++;
              }
          }
      }

      if ( inlierCount > bestinlierCount )
      {
          for (int n = 0; n < 9; n++)
          {
              hbest[n] = float_h[n];
          }
          best = rounds-1;
          bestinlierCount = inlierCount;
      }
  }

  CvMat * h_output = cvCreateMat( 3, 3, CV_32F );
  for (int i = 0; i < 3; ++i)
      for (int j = 0; j < 3; ++j)
          CV_MAT_ELEM(*h_output, float, i, j) = hbest[i*3+j];

  return h_output;

}

void computeHomography ( const std::vector<Feature> &      f1,
                         const std::vector<Feature> &      f2,
                         const std::vector<FeatureMatch> & matches,
                         CvMat *                           h )
{
    float x1[4];
    float y1[4];
    float x2[4];
    float y2[4];

    for (int i = 0; i < 4; ++i)
    {
        x1[i] = (float)f1[matches[i].id1-1].x;
        y1[i] = (float)f1[matches[i].id1-1].y;
        x2[i] = (float)f2[matches[i].id2-1].x;
        y2[i] = (float)f2[matches[i].id2-1].y;
    }

    float Am[] = {    x2[0],  y2[0],  1.f,      0.f,      0.f,      0.f, -x1[0]*x2[0], -x1[0]*y2[0],
                      0.f,    0.f,    0.f,      x2[0],    y2[0],    1.f, -y1[0]*x2[0], -y1[0]*y2[0],
                      x2[1],  y2[1],  1.f,      0.f,      0.f,      0.f, -x1[1]*x2[1], -x1[1]*y2[1],
                      0.f,    0.f,    0.f,      x2[1],    y2[1],    1.f, -y1[1]*x2[1], -y1[1]*y2[1],
                      x2[2],  y2[2],  1.f,      0.f,      0.f,      0.f, -x1[2]*x2[2], -x1[2]*y2[2],
                      0.f,    0.f,    0.f,      x2[2],    y2[2],    1.f, -y1[2]*x2[2], -y1[2]*y2[2],
                      x2[3],  y2[3],  1.f,      0.f,      0.f,      0.f, -x1[3]*x2[3], -x1[3]*y2[3],
                      0.f,    0.f,    0.f,      x2[3],    y2[3],    1.f, -y1[3]*x2[3], -y1[3]*y2[3] };

    float Bm[] = { x1[0], y1[0], x1[1], y1[1], x1[2], y1[2], x1[3], y1[3] };

  cv::Mat A( 8, 8, CV_32F, Am );
  cv::Mat h_mat(8, 1, CV_32F);

  cv::Mat B(8,1,CV_32F, Bm);
  cv::solve(A,B,h_mat,cv::DECOMP_SVD);

  cv::Mat h_mat1 = h;


  for (int i = 0; i < 9; ++i)
  {
      if (i != 8)
        h_mat1.at<float>(i,0) = h_mat.at<float>(i,0);
      else
        h_mat1.at<float>(i,0) = 1.f;
  }
}

void applyHomography ( float   x,
                       float   y,
                       float & xNew,
                       float & yNew,
                       float   h [9] )
{

  float d = h[6] * x + h[7] * y + h[8];

  xNew = ( h[0] * x + h[1] * y + h[2] ) / d;
  yNew = ( h[3] * x + h[4] * y + h[5] ) / d;
}