/*
 * Processor.h
 *
 *  Created on: Jun 13, 2010
 *      Author: ethan
 */

#ifndef PROCESSOR_H_
#define PROCESSOR_H_

#include <opencv2/core/core.hpp>
#include <opencv2/features2d/features2d.hpp>
//#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/video/tracking.hpp>


#include <vector>
#include <time.h>
#include <math.h>
#include "image_pool.h"
#include "FeatureSet.h"
//#include "OpticalFlow.h"

#define DETECT_FAST 0
#define DETECT_STAR 1
#define DETECT_SURF 2

class Processor {

	cv::StarFeatureDetector stard;
	cv::FastFeatureDetector fastd;
	cv::SurfFeatureDetector surfd;
	std::vector<cv::KeyPoint> keypoints;

	vector<vector<cv::Point2f> > imagepoints;

	cv::Mat K;
	cv::Mat distortion;
	cv::Size imgsize;
	
	cv::Mat H;
	vector<cv::Point2f> accum_offset;
	
	//image_pool pool;
public:

	Processor();
	virtual ~Processor();

	void detectAndDrawFeatures(int idx, image_pool* pool, int feature_type);

	bool detectAndDrawChessboard(int idx,image_pool* pool);

	void resetChess();

	int getNumberDetectedChessboards();

	void calibrate(const char* filename);

	void drawText(int idx, image_pool* pool, const char* text);

	void opticalFlowInit(image_pool* pool);

	void opticalFlow(image_pool* pool, int* capture, int* distx, int* disty);

	void saveImage(int input_idx,image_pool* pool, char* BYTE);
	
	void captureImage(int input_idx, image_pool* pool, int output_idx);

	void setSphericalCoord(float f, float s);

	void getSphericalImage(char* output);
	
	void makePanorama1(char* img1c, char* img2c, image_pool* pool, int* width, int* height);

	void makePanorama2(char* img1c, image_pool* pool, int width, int height);
	
	void makePanoUseOpticalFlow1(char* img1c, char* img2c, int* width, int* height);

	void makePanoUseOpticalFlow2(char* img, int width, int height);

};
//
//class OpticalFlow
//{
//public:
//	OpticalFlow(void);
//	virtual ~OpticalFlow(void);
//
//	void InitInstance(CvSize size);
//	void ExitInstance();
//
//	int InitFlow(IplImage *image);
//	int GetFlow2(IplImage *image, CvPoint2D32f *displacement_vector, CvPoint2D32f *velocity_vector, float *orientation);
//	void ResetFlow();
//	void DrawFlow(IplImage *out_image);
//	inline float deg2rad(float deg);
//	inline float rad2deg(float rad);
//	inline int get_distribution(std::vector<float>&input, float *mean, float *var);
//	int GetFASTFeatures(IplImage *image, CvPoint2D32f *pts);
//	void ComputeFASTFeatures ( IplImage * image, FeatureSet & features );
//	
//private:
//	IplImage *curr_image;
//	IplImage *prev_image;
//
//	IplImage *eig_image;
//	IplImage *tmp_image;
//
//	IplImage* pyr_prev;	//prev_image
//	IplImage* pyr_curr;	//curr_image
//
//	CvPoint2D32f *corners_prev;
//	CvPoint2D32f *corners_curr;
//	CvPoint2D32f *corners_output;
//	CvSize winsize;
//
//	time_t currtime;
//	time_t prevtime;
//	time_t starttime;
//	double deltatime;
//
//	int corner_count_prev;
//	int corner_count_curr;
//	typedef struct _matches matches_t;
//	std::vector<matches_t> matches;
//
//	int RansacFlow(std::vector<matches_t> &m, CvPoint2D32f *displacement, float *orientation);
//};

// FeatureMatch is used by your feature matching routine to store the
// ID of the matching feature in the other image, as well as the score
// of the match.
//struct FeatureMatch
//{
//  size_t id1, id2;
//  float  score;
//};
//
//// The Feature class stores the feature ID, location, and a vector of
//// whatever attributes you choose to use.  It also has methods for
//// drawing the feature and printing its description to the console.
//// Feel free to change these.
//class Feature {
//public:
//int type;
//size_t id;
//int x;
//int y;
//float angleRadians;
//
//vector<float> data;
//
//bool selected;
//
//public:
//// Create a feature.
//Feature ();
//
//};
//
//// The FeatureSet class represents a vector of features for a single
//// image.  You don't need to modify it.
//class FeatureSet : public vector<Feature> {
//public:
//	// Create a feature set.
//	FeatureSet();
//};


#endif /* PROCESSOR_H_ */
