/*
 * include the headers required by the generated cpp code
 */
%{

#include "Processor.h"
#include "image_pool.h"

using namespace cv;
%}
%include "typemaps.i"
%include "arrays_java.i"
%include "various.i"
%apply int *OUTPUT {int *width};
%apply int *OUTPUT {int *height};
%apply int *OUTPUT {int *capture};
%apply int *OUTPUT {int *distx};
%apply int *OUTPUT {int *disty};
%apply char *BYTE {char *img1c};
%apply char *BYTE {char *img2c};
%apply char *BYTE {char *output};
%apply char *BYTE {char *img};

/**
 * some constants, see Processor.h
 */
#define DETECT_FAST 0
#define DETECT_STAR 1
#define DETECT_SURF 2

//import the android-cv.i file so that swig is aware of all that has been previous defined
//notice that it is not an include....
%import "android-cv.i"

//make sure to import the image_pool as it is 
//referenced by the Processor java generated
//class
%typemap(javaimports) Processor "
import com.opencv.jni.image_pool;// import the image_pool interface for playing nice with
								 // android-opencv

/** Processor - for processing images that are stored in an image pool
*/"

class Processor {
public:
	Processor();
	virtual ~Processor();

	
	
	void detectAndDrawFeatures(int idx, image_pool* pool, int feature_type);

	bool detectAndDrawChessboard(int idx,image_pool* pool);
	
	void resetChess();
	
	int getNumberDetectedChessboards();
	
	void calibrate(const char* filename);
	
	void drawText(int idx, image_pool* pool, const char* text);
	
	void opticalFlowInit(image_pool* pool);

	void opticalFlow(image_pool* pool, int* capture, int* distx, int* disty);
	
	void saveImage(int input_idx,image_pool* pool, char* BYTE);
	
	void captureImage(int input_idx, image_pool* pool, int output_idx);
	
	void setSphericalCoord(float f, float s);
	
	void getSphericalImage(char* output);
	
	void makePanorama1(char* img1c, char* img2c, image_pool* pool, int* width, int* height);
	
	void makePanorama2(char* img1c, image_pool* pool, int width, int height);
	
	void makePanoUseOpticalFlow1(char* img1c, char* img2c, int* width, int* height);
	
	void makePanoUseOpticalFlow2(char* img, int width, int height);
	
};

