package com.theveganrobot.cvcamera;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.app.ProgressDialog;
import android.content.DialogInterface;
import android.content.pm.ActivityInfo;
import android.graphics.Bitmap;
import android.graphics.Bitmap.CompressFormat;
import android.graphics.Color;
import android.os.Bundle;
import android.os.Environment;
import android.util.Log;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup.LayoutParams;
import android.view.Window;
import android.view.WindowManager;
import android.widget.Button;
import android.widget.FrameLayout;
import android.widget.Gallery;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.Toast;

import com.opencv.camera.NativePreviewer;
import com.opencv.camera.NativeProcessor;
import com.opencv.camera.NativeProcessor.PoolCallback;
import com.opencv.jni.image_pool;
import com.opencv.opengl.GL2CameraViewer;
import com.theveganrobot.cvcamera.jni.Processor;
import com.theveganrobot.cvcamera.jni.cvcamera;

public class CVCamera extends Activity {

	static final int IMG_H = 480;
	static final int IMG_W = 640;
	static final int DIALOG_CALIBRATING = 0;
	static final int DIALOG_CALIBRATION_FILE = 1;
	private static final int DIALOG_OPENING_TUTORIAL = 2;
	private static final int DIALOG_TUTORIAL_FAST = 3;
	private static final int DIALOG_TUTORIAL_SURF = 4;
	private static final int DIALOG_TUTORIAL_STAR = 5;
	private static final int DIALOG_TUTORIAL_CHESS = 6;
	private static final int DIALOG_SAVING_IMAGE = 7;
	private static final int DIALOG_IMAGE_SAVED = 8;
	private static final int DIALOG_IMAGE_CAPTURING = 9;
	private static final int DIALOG_IMAGE_CAPTURED = 10;
	private static final int DIALOG_IMAGE_PANORAMA = 11;
	private static final int DIALOG_IMAGE_PANORAMA_DONE = 12;
	private static final int DIALOG_IMAGE_SAVED3 = 13;
	private boolean captureChess;

	ProgressDialog makeCalibDialog() {
		ProgressDialog progressDialog;
		progressDialog = new ProgressDialog(this);
		progressDialog.setMessage("Callibrating. Please wait...");
		progressDialog.setCancelable(false);

		return progressDialog;
	}

	void toasts(int id) {
		switch (id) {
		case DIALOG_OPENING_TUTORIAL:
			Toast.makeText(this, "Try clicking the menu for CV options.",
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_TUTORIAL_FAST:
			Toast.makeText(this, "Detecting and Displaying FAST features",
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_TUTORIAL_SURF:
			Toast.makeText(this, "Detecting and Displaying SURF features",
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_TUTORIAL_STAR:
			Toast.makeText(this, "Detecting and Displaying STAR features",
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_TUTORIAL_CHESS:
			Toast
					.makeText(
							this,
							"Calibration Mode, Point at a chessboard pattern and press the camera button, space,"
									+ "or the DPAD to capture.",
							Toast.LENGTH_LONG).show();
			break;
		case DIALOG_SAVING_IMAGE:
			Toast.makeText(this, "Saving image...", 
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_IMAGE_SAVED:
			Toast.makeText(this, "Image saved successfully", 
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_IMAGE_CAPTURING:
			Toast.makeText(this, "Capturing image...", 
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_IMAGE_CAPTURED:
			Toast.makeText(this, "Image captured! "+capture_idx, 
					Toast.LENGTH_SHORT).show();
			break;
		case DIALOG_IMAGE_PANORAMA:
			Toast.makeText(this, "Making panorama",  
					Toast.LENGTH_LONG).show();
			break;
		case DIALOG_IMAGE_PANORAMA_DONE:
			Toast.makeText(this, "Making panorama DONE!",  
					Toast.LENGTH_LONG).show();
			break;	
		case DIALOG_IMAGE_SAVED3:
			Toast.makeText(this, "Image saving stage3",  
					Toast.LENGTH_SHORT).show();
			break;
		default:
			break;
		}

	}

	@Override
	protected Dialog onCreateDialog(int id) {
		Dialog dialog;
		switch (id) {
		case DIALOG_CALIBRATING:
			dialog = makeCalibDialog();
			break;

		case DIALOG_CALIBRATION_FILE:
			dialog = makeCalibFileAlert();
			break;
			
		case DIALOG_IMAGE_CAPTURING:
			dialog = makeImgGrabDialog();
			break;
			
		case DIALOG_IMAGE_PANORAMA:
			dialog = makePanoDialog();
			break;
		default:
			dialog = null;
		}
		return dialog;
	}

	private Dialog makeCalibFileAlert() {
		AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setMessage(calib_text).setTitle(
				"camera.yml at " + calib_file_loc).setCancelable(false)
				.setPositiveButton("Ok", new DialogInterface.OnClickListener() {
					public void onClick(DialogInterface dialog, int id) {

					}
				});
		AlertDialog alert = builder.create();
		return alert;
	}
	private Dialog makeImgGrabDialog() {
		ProgressDialog dialog = ProgressDialog.show(CVCamera.this, "", 
                "Capturing. Please wait...", true);
		return dialog;
	}
	private Dialog makePanoDialog() {
		ProgressDialog dialog = ProgressDialog.show(CVCamera.this, "", 
                "Making panorama. Please wait...", true);
		return dialog;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onKeyUp(int, android.view.KeyEvent)
	 */
	@Override
	public boolean onKeyUp(int keyCode, KeyEvent event) {

		switch (keyCode) {
		case KeyEvent.KEYCODE_CAMERA:
		case KeyEvent.KEYCODE_SPACE:
		case KeyEvent.KEYCODE_DPAD_CENTER:
			captureChess = true;
			return true;

		default:
			return super.onKeyUp(keyCode, event);
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onKeyLongPress(int, android.view.KeyEvent)
	 */
	@Override
	public boolean onKeyLongPress(int keyCode, KeyEvent event) {

		return super.onKeyLongPress(keyCode, event);
	}

	/**
	 * Avoid that the screen get's turned off by the system.
	 */
	public void disableScreenTurnOff() {
		getWindow().setFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON,
				WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);
	}

	/**
	 * Set's the orientation to landscape, as this is needed by AndAR.
	 */
	public void setOrientation() {
		setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE);
	}

	/**
	 * Maximize the application.
	 */
	public void setFullscreen() {
		requestWindowFeature(Window.FEATURE_NO_TITLE);
		getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,
				WindowManager.LayoutParams.FLAG_FULLSCREEN);
	}

	public void setNoTitle() {
		requestWindowFeature(Window.FEATURE_NO_TITLE);
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		menu.add("FAST");
		menu.add("STAR");
		menu.add("SURF");
		menu.add("Chess");
		menu.add("saveImage");
		menu.add("Capture");
		menu.add("Panorama");
		return true;
	}

	private NativePreviewer mPreview;
	private GL2CameraViewer glview;

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		LinkedList<PoolCallback> defaultcallbackstack = new LinkedList<PoolCallback>();
		defaultcallbackstack.addFirst(glview.getDrawCallback());
		if (item.getTitle().equals("FAST")) {

			defaultcallbackstack.addFirst(new FastProcessor());
			toasts(DIALOG_TUTORIAL_FAST);
		}

		else if (item.getTitle().equals("Chess")) {

			defaultcallbackstack.addFirst(new CalibrationProcessor());
			toasts(DIALOG_TUTORIAL_CHESS);

		}

		else if (item.getTitle().equals("STAR")) {

			defaultcallbackstack.addFirst(new STARProcessor());
			toasts(DIALOG_TUTORIAL_STAR);

		}

		else if (item.getTitle().equals("SURF")) {

			defaultcallbackstack.addFirst(new SURFProcessor());
			toasts(DIALOG_TUTORIAL_SURF);

		}
		
		else if (item.getTitle().equals("saveImage")) {

			defaultcallbackstack.addFirst(new saveImageProcessor());
			toasts(DIALOG_SAVING_IMAGE);

		}
		else if (item.getTitle().equals("Capture")) {

			defaultcallbackstack.addFirst(new CaptureProcessor());
			//toasts(DIALOG_IMAGE_CAPTURING);
			runOnUiThread(new Runnable() {

				@Override
				public void run() {
					showDialog(DIALOG_IMAGE_CAPTURING);

				}
			});

		}
		else if (item.getTitle().equals("Panorama")) {
			defaultcallbackstack.addFirst(new OpticalFlowProcessor());
			defaultcallbackstack.addFirst(new PanoramaProcessor());
//			runOnUiThread(new Runnable() {
//
//				@Override
//				public void run() {
//					showDialog(DIALOG_IMAGE_PANORAMA);
//
//				}
//			});

		}


		mPreview.addCallbackStack(defaultcallbackstack);
		return true;
	}

	@Override
	public void onOptionsMenuClosed(Menu menu) {
		// TODO Auto-generated method stub
		super.onOptionsMenuClosed(menu);
	}

	FrameLayout frame;
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		setFullscreen();
		disableScreenTurnOff();

		frame = new FrameLayout(this);

		// Create our Preview view and set it as the content of our activity.
		mPreview = new NativePreviewer(getApplication(), IMG_W, IMG_H);

		LayoutParams params = new LayoutParams(LayoutParams.WRAP_CONTENT,
				LayoutParams.WRAP_CONTENT);
		params.height = getWindowManager().getDefaultDisplay().getHeight();
		params.width = (int) (params.height * 4.0 / 2.88);

		LinearLayout vidlay = new LinearLayout(getApplication());

		vidlay.setGravity(Gravity.CENTER);
		vidlay.addView(mPreview, params);
		frame.addView(vidlay);

		// make the glview overlay ontop of video preview
		mPreview.setZOrderMediaOverlay(false);

		glview = new GL2CameraViewer(getApplication(), false, 0, 0);
		glview.setZOrderMediaOverlay(true);
		glview.setLayoutParams(new LayoutParams(160,
				120));
		
		ImageButton capture_button = new ImageButton(getApplicationContext());
		capture_button.setImageDrawable(getResources().getDrawable(android.R.drawable.ic_menu_camera));
		capture_button.setLayoutParams(new LayoutParams(LayoutParams.WRAP_CONTENT,
				LayoutParams.WRAP_CONTENT));
		capture_button.setOnClickListener(new View.OnClickListener() {
			
			
			@Override
			public void onClick(View v) {
				captureChess = true;
				
			}
		});
		
		LinearLayout buttons = new LinearLayout(getApplicationContext());
		buttons.setLayoutParams(new LayoutParams(LayoutParams.WRAP_CONTENT,
				LayoutParams.WRAP_CONTENT));
	
		buttons.addView(capture_button);
		
		Button focus_button = new Button(getApplicationContext());
		focus_button.setLayoutParams(new LayoutParams(LayoutParams.WRAP_CONTENT,
				LayoutParams.WRAP_CONTENT));
		focus_button.setText("Focus");
		focus_button.setOnClickListener(new View.OnClickListener() {
			
			@Override
			public void onClick(View v) {
				mPreview.postautofocus(100);
			}
		});
		buttons.addView(focus_button);
		
		LinearLayout glay = new LinearLayout(getApplication());
		glay.setGravity(Gravity.RIGHT);
		glay.addView(glview);
		frame.addView(glay);

		frame.addView(buttons);
		setContentView(frame);
		toasts(DIALOG_OPENING_TUTORIAL);

	}

	@Override
	public boolean onTrackballEvent(MotionEvent event) {
		if(event.getAction() == MotionEvent.ACTION_UP){
			captureChess = true;
			return true;
		}
		return super.onTrackballEvent(event);
	}

	@Override
	protected void onPause() {
		super.onPause();

		
		//clears the callback stack
		mPreview.onPause();

		glview.onPause();

	}

	@Override
	protected void onResume() {
		super.onResume();
		glview.onResume();
		
		//add an initiall callback stack to the preview on resume...
		//this one will just draw the frames to opengl
		LinkedList<NativeProcessor.PoolCallback> cbstack = new LinkedList<PoolCallback>();
		cbstack.add(glview.getDrawCallback());
		mPreview.addCallbackStack(cbstack);
		mPreview.onResume();

	}

	// final processor so that these processor callbacks can access it
	final Processor processor = new Processor();

	class FastProcessor implements NativeProcessor.PoolCallback {

		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			processor.detectAndDrawFeatures(idx, pool, cvcamera.DETECT_FAST);

		}

	}

	class STARProcessor implements NativeProcessor.PoolCallback {

		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			processor.detectAndDrawFeatures(idx, pool, cvcamera.DETECT_STAR);
						

		}

	}

	class SURFProcessor implements NativeProcessor.PoolCallback {

		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			processor.detectAndDrawFeatures(idx, pool, cvcamera.DETECT_SURF);

		}

	}
	
	int capture_idx = 0;
	List<byte[]> imglist = new LinkedList<byte[]>();
	List<int[]> imgsphdist = new LinkedList<int[]>();
	
	class CaptureProcessor implements NativeProcessor.PoolCallback {
		boolean captured = false;
		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			if (captured)
			{
				return;
			}
			
			byte[] img = new byte[IMG_W*IMG_H*3];
			processor.saveImage(0, pool, img);
//			saveImage2SD(img, IMG_W, IMG_H, 90);
			imglist.add(img);
			capture_idx++;
			captured = true;
			runOnUiThread(new Runnable() {
				@Override
				public void run() {
					removeDialog(DIALOG_IMAGE_CAPTURING);
					toasts(DIALOG_IMAGE_CAPTURED);
				}
			});
		}
	}
	
	boolean do_pano = false;
	class PanoramaProcessor implements NativeProcessor.PoolCallback {

		boolean saved = false;
		int oldimgsize = 0;
		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
//			if (imglist.size() > oldimgsize)
//			{
//				// making spherical image in background process
//				processor.getSphericalImage(oldimgsize);
//				oldimgsize = imglist.size();
//			}
			
			if (saved || do_pano == false)
			{
				return;
			}
			runOnUiThread(new Runnable() {
				@Override
				public void run() {
					showDialog(DIALOG_IMAGE_PANORAMA);
				}
			});
			
			int[] width = new int[1];
			int[] height = new int[1];
			for (int i = 0; i < 2; ++i)
				processor.getSphericalImage(imglist.get(i));
//			processor.makePanorama1(imglist.get(0), imglist.get(1), pool, width, height);
//			byte[] img = new byte[width[0]*height[0]*3];
//			processor.makePanorama2(img, pool, width[0], height[0]);
			
			processor.makePanoUseOpticalFlow1(imglist.get(0), imglist.get(1), width, height);
			byte[] img = new byte[width[0]*height[0]*3];
			processor.makePanoUseOpticalFlow2(img, width[0], height[0]);
			
			saveImage2SD(img, width[0], height[0], 100);
			saved = true;
			runOnUiThread(new Runnable() {
				@Override
				public void run() {
					
					removeDialog(DIALOG_IMAGE_PANORAMA);
					toasts(DIALOG_IMAGE_PANORAMA_DONE);
					final ImageView imV = new ImageView(getApplicationContext());
					imV.setImageBitmap(myImage);
					imV.setLayoutParams(new Gallery.LayoutParams(IMG_W, IMG_H));
					imV.setOnClickListener(new View.OnClickListener() {
						@Override
						public void onClick(View v) {
							frame.removeView(imV);
						}
					});
					
					frame.addView(imV);
				}
			});
		}

	}
	int maxdisty = 0;
	class OpticalFlowProcessor implements NativeProcessor.PoolCallback {
		boolean OFinit = false;
		int[] c = new int[1];
		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			if (do_pano == true)
			{
				return;
			}
			if (OFinit == false)
			{
				processor.opticalFlowInit(pool);
				OFinit = true;
			}
			else
			{
				int[] distx = new int[1];
				int[] disty = new int[1];
				processor.opticalFlow(pool, c, distx, disty);
				Log.d("OpticalFlow", "OpticalFlow dist x: " + distx[0] + " y: " + disty[0]);
				if (c[0] == 1)
				{
					c[0] = 0;
					byte[] img = new byte[IMG_W*IMG_H*3];
					processor.saveImage(0, pool, img);
					imglist.add(img);
					imgsphdist.add(distx);
					if (maxdisty < disty[0])
						maxdisty = disty[0];
					capture_idx++;
					Log.d("OpticalFlow", "OpticalFlow got image " + capture_idx);
				
					runOnUiThread(new Runnable() {
						@Override
						public void run() {
							toasts(DIALOG_IMAGE_CAPTURED);
						}
					});
				}
				if (capture_idx > 2)
				{
					do_pano = true;
				}
					
			}
		}
		
	}
	
	String save_file_loc = null;
	int savedImageN = 1;
	Bitmap myImage;
	public void saveImage2SD(byte[] imgbytes, int width, int height, int quality){
		File opencvdir = new File(Environment
				.getExternalStorageDirectory(), "opencv");
		if (!opencvdir.exists()) {
			opencvdir.mkdir();
		}
		
		File imgfile = new File(opencvdir, "savedImage"+savedImageN+".png");
		
		FileOutputStream fileOutputStream = null;
		save_file_loc = imgfile.getAbsolutePath();
		int[] colors = new int[width*height];
		int r = 0;
		int g = 0;
		int b = 0;
		int a = 0;
		for (int i = 0; i < colors.length; ++i)
		{
			r = (imgbytes[i*3]+256)%256;// < 0 ? imgbytes[i*3]+256: imgbytes[i*3];
			g = (imgbytes[(i*3+1)]+256)%256;// <0 ? imgbytes[(i*3+1)]+256: imgbytes[(i*3+1)];
			b = (imgbytes[(i*3+2)]+256)%256;// <0 ? imgbytes[(i*3+2)]+256: imgbytes[(i*3+2)];
//			if (r != 0 && g != 0 && b != 0)
				a = 255;
//			else
//				a = 0;
			
			colors[i] = Color.argb(a, r, g, b);
		}
		myImage = Bitmap.createBitmap(colors, width, height, Bitmap.Config.ARGB_8888);

		try {
			fileOutputStream = new FileOutputStream(save_file_loc);
			
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		BufferedOutputStream bos = new BufferedOutputStream(fileOutputStream);
		myImage.compress(CompressFormat.PNG, quality, bos);
		try {
			bos.flush();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		try {
			bos.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}		
		savedImageN++;
	}
	
	class saveImageProcessor implements NativeProcessor.PoolCallback {

		boolean saved = false;
		
		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {
			if (saved)
			{
//				toasts(DIALOG_IMAGE_SAVED);
				return;
			}
			
			byte[] imgbytes = new byte[IMG_W*IMG_H*3];
						
			processor.saveImage(idx, pool, imgbytes);
			saveImage2SD(imgbytes, IMG_W, IMG_H, 90);
			
			saved = true;
			
		}
		
	}

	String calib_text = null;
	String calib_file_loc = null;

	class CalibrationProcessor implements NativeProcessor.PoolCallback {

		boolean calibrated = false;

		@Override
		public void process(int idx, image_pool pool, long timestamp,
				NativeProcessor nativeProcessor) {

			if (calibrated) {
				processor.drawText(idx, pool, "Calibrated successfully");
				return;
			}
			if (processor.getNumberDetectedChessboards() == 10) {

				File opencvdir = new File(Environment
						.getExternalStorageDirectory(), "opencv");
				if (!opencvdir.exists()) {
					opencvdir.mkdir();
				}
				File calibfile = new File(opencvdir, "camera.yml");

				calib_file_loc = calibfile.getAbsolutePath();
				processor.calibrate(calibfile.getAbsolutePath());
				Log.i("chessboard", "calibrated");
				calibrated = true;
				processor.resetChess();
				runOnUiThread(new Runnable() {

					@Override
					public void run() {
						removeDialog(DIALOG_CALIBRATING);

					}
				});

				try {

					StringBuilder text = new StringBuilder();
					String NL = System.getProperty("line.separator");
					Scanner scanner = new Scanner(calibfile);

					try {
						while (scanner.hasNextLine()) {
							text.append(scanner.nextLine() + NL);
						}
					} finally {
						scanner.close();
					}

					calib_text = text.toString();

					runOnUiThread(new Runnable() {

						@Override
						public void run() {
							showDialog(DIALOG_CALIBRATION_FILE);

						}
					});

				} catch (FileNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			} else if (captureChess
					&& processor.detectAndDrawChessboard(idx, pool)) {

				runOnUiThread(new Runnable() {

					String numchess = String.valueOf(processor
							.getNumberDetectedChessboards());

					@Override
					public void run() {
						Toast.makeText(CVCamera.this,
								"Detected " + numchess + " of 10 chessboards",
								Toast.LENGTH_SHORT).show();

					}
				});
				Log.i("cvcamera",
						"detected a chessboard, n chess boards found: "
								+ String.valueOf(processor
										.getNumberDetectedChessboards()));

			}

			captureChess = false;

			if (processor.getNumberDetectedChessboards() == 10) {
				runOnUiThread(new Runnable() {

					@Override
					public void run() {
						showDialog(DIALOG_CALIBRATING);

					}
				});

				processor.drawText(idx, pool, "Calibrating, please wait.");
			}
			if (processor.getNumberDetectedChessboards() < 10) {

				processor.drawText(idx, pool, "found "
						+ processor.getNumberDetectedChessboards()
						+ "/10 chessboards");
			}

		}

	}

}
